#include "Interpreter.h"

t_key keys[MAX_KEYS];
t_dinput dinputs[MAX_DINPUTS];
t_ainput ainputs[MAX_AINPUTS];
t_statemachine statemachines[MAX_STATEMACHINES];
t_relay relays[MAX_RELAYS];
CODERAMROM unsigned char code[MAX_CODE];

t_codeheader* pcodeheader;
unsigned char* pcodebytes;


/**********************************************************
Alle Strukturen initialisieren
**********************************************************/
void InitInterpreter(void){
int i;

  pcodeheader=(t_codeheader*)&code[0];
  pcodebytes=&pcodeheader->bytes[0];

  for (i=0;i<MAX_KEYS;i++){
    keys[i].pressed=0;
    keys[i].released=0;
    keys[i].time=0;
    keys[i].down=0;
    keys[i].clickdetected=0;
  }

  for (i=0;i<MAX_DINPUTS;i++){
    dinputs[i].lh=0;
    dinputs[i].hl=0;
    dinputs[i].high=0;
    dinputs[i].old=0;
  }

  for (i=0;i<MAX_AINPUTS;i++){
    ainputs[i].val=0;
  }

  for (i=0;i<MAX_STATEMACHINES;i++){
    statemachines[i].state=pcodeheader->initstate[i];
    statemachines[i].laststate=0;
    statemachines[i].has_changed=0;
    statemachines[i].time=0;
  }

  for (i=0;i<MAX_RELAYS;i++){
    relays[i].active=0;
    relays[i].old=0;
    relays[i].breakbmake=0;
  }
}


/**********************************************************
Ist aufzurufen, wenn eine Taste gedrueckt wurde
**********************************************************/
void Key_Down(int i){
  keys[i].time=1;
  keys[i].pressed=1;
  keys[i].down=1;
}

/**********************************************************
Ist aufzurufen, wenn eine Taste losgelassen wurde
**********************************************************/
void Key_Up(int i){

  if (keys[i].clickdetected)
    keys[i].clickdetected=0;
  else
    keys[i].released=1;

  keys[i].down=0; 
}

/**********************************************************
Werte der Digitaleingaenge uebernehmen
**********************************************************/
void SetDInput(int i,unsigned char val){
  dinputs[i].high=val;
  if (dinputs[i].old!=val){
    if (dinputs[i].old==0)
      dinputs[i].lh=1;
    else
      dinputs[i].hl=1;
  }
  dinputs[i].old=val;
}

/**********************************************************
Werte der Analogeingaenge uebernehmen
**********************************************************/
void SetAInput(int i,unsigned char val){
  ainputs[i].val=val;
}


/**********************************************************
Tasteneingaenge behandeln
**********************************************************/
void Keys(){
int i;
  for (i=0;i<MAX_KEYS;i++){
    keys[i].pressed=0;
    keys[i].released=0;
    if (keys[i].time<255)
      keys[i].time++;
  }
}

/**********************************************************
Digitaleingaenge behandeln
**********************************************************/
void DInputs(){
int i;
  for (i=0;i<MAX_DINPUTS;i++){
    dinputs[i].lh=0;
    dinputs[i].hl=0;
  }
}


/**********************************************************
Zustandsmaschinen behandeln
**********************************************************/
void Statemachines(){
int i;
  for (i=0;i<pcodeheader->statemachines;i++){
    if(statemachines[i].has_changed){
      statemachines[i].time=0;
      statemachines[i].has_changed=0;
    }
    if (statemachines[i].time<60000)
      statemachines[i].time++;
  }
}


/**********************************************************
Ausgaenge behandeln
**********************************************************/
void RelOutputs(){
int i;
  for (i=0;i<MAX_RELAYS;i++){
     relays[i].breakbmake= (relays[i].active) && (relays[i].old);
     relays[i].old=relays[i].active;
     SetRelais(i,relays[i].breakbmake);
  }
}

/**********************************************************
50ms Task. Erzeugt 2 phasenversetzte 100ms Tasks
**********************************************************/
void Task50ms(void){
static unsigned char taskcnt=0;
  taskcnt=!taskcnt;
  if (taskcnt)
    Task100ms_A();
  else
    Task100ms_B();
}

/**********************************************************
1. 100ms Task. Wird benutzt, um Relais mit 
Break before Make Verhalten anzusteuern
**********************************************************/
void Task100ms_A(void){
  RelOutputs();
}

/**********************************************************
2. 100ms Task. Interpretiert den Bytecode und ruft alle
Funktionen zur Behandlung der Ein-und Ausgaenge auf
**********************************************************/
void Task100ms_B(void){

  Interpret();

  Keys();
  DInputs();
  Statemachines();
  RelOutputs();
}

/**********************************************************
Bytecode Interpreter
**********************************************************/
void Interpret(void){
t_actions* paction;
t_or_compares* porcompare;
t_compares* pcompare;
int offset,or_offset,and_offset;
unsigned char or_cond,and_cond;
int and_i,or_i;
short int compvalue;
int i;
unsigned short int commands;
unsigned char keyclicknr;
unsigned char keyclickdetected;

  offset=0;
  // alle Kommandos abarbeiten
  for (commands=0;commands<BIGENDIAN(pcodeheader->length);commands++){
    paction=(t_actions*)(pcodebytes+offset);
    or_offset=0;
    or_cond=0;
    // alle ODER-verknuepften Bedingungen innerhalb eines Kommandos abarbeiten
    for (or_i=0;or_i<paction->ors;or_i++){
      porcompare=(t_or_compares*)((unsigned char*)&paction->orcompares[0]+or_offset);
      and_cond=!or_cond;
      and_offset=0;
      keyclickdetected=0;
      // alle UND-verknuepften Bedingungen innerhalb eines Kommandos abarbeiten
      for (and_i=0;and_i<porcompare->ands;and_i++){
        if (and_cond){
          pcompare=(t_compares*)((unsigned char*)&porcompare->andcompares[0]+and_offset);
          
          // Je nach TYP der Bedingung den entsprechenden Wert zum Vergleich holen
          if (pcompare->object.obj_typ==KEYCLICK){
            compvalue=keys[pcompare->object.obj_nr].pressed;
            keyclickdetected=1;
            keyclicknr=pcompare->object.obj_nr;
          }
          if (pcompare->object.obj_typ==KEYPRESSED){
            compvalue=keys[pcompare->object.obj_nr].pressed;
          }
          if (pcompare->object.obj_typ==KEYRELEASED){
            compvalue=keys[pcompare->object.obj_nr].released;
          }
          if (pcompare->object.obj_typ==KEYRELEASETIME){
            compvalue=keys[pcompare->object.obj_nr].time;
            if (!keys[pcompare->object.obj_nr].released)
              and_cond=0;
          }
          if (pcompare->object.obj_typ==KEYDOWN){
            compvalue=keys[pcompare->object.obj_nr].down;
          }
          if (pcompare->object.obj_typ==KEYDOWNTIME){
            compvalue=keys[pcompare->object.obj_nr].time;
            if (!keys[pcompare->object.obj_nr].down)
              and_cond=0;
          }
          if (pcompare->object.obj_typ==KEYUP){
            compvalue=!keys[pcompare->object.obj_nr].down;
          }

          if (pcompare->object.obj_typ==DINPUTHIGH){
            compvalue=dinputs[pcompare->object.obj_nr].high;
          }
          if (pcompare->object.obj_typ==DINPUTLOW){
            compvalue=!dinputs[pcompare->object.obj_nr].high;
          }
          if (pcompare->object.obj_typ==DINPUTLH){
            compvalue=dinputs[pcompare->object.obj_nr].lh;
          }
          if (pcompare->object.obj_typ==DINPUTHL){
            compvalue=dinputs[pcompare->object.obj_nr].hl;
          }

          if (pcompare->object.obj_typ==AINPUT){
            compvalue=ainputs[pcompare->object.obj_nr].val;
          }

          if (pcompare->object.obj_typ==RELAY){
            compvalue=relays[pcompare->object.obj_nr].active;
          }

          if (pcompare->object.obj_typ==STATE){
            compvalue=statemachines[pcompare->object.obj_nr].state;
          }
          if (pcompare->object.obj_typ==STATETIME){
            compvalue=statemachines[pcompare->object.obj_nr].time;
          }



          // Je nach Art des Vergleichsoperators den Wert mit dem zu vergleichenden Wert vergleichen
          if (pcompare->comptyp==NEVER){
            and_cond=0;
          }
          if (pcompare->comptyp==EQUAL){
             if (!(compvalue==BIGENDIAN(pcompare->compvalue)))
               and_cond=0;
          }
          if (pcompare->comptyp==UNEQUAL){
             if (!(compvalue!=BIGENDIAN(pcompare->compvalue)))
               and_cond=0;
          }
          if (pcompare->comptyp==GREATER){
             if (!(compvalue>BIGENDIAN(pcompare->compvalue)))
               and_cond=0;
          }
          if (pcompare->comptyp==SMALLER){
             if (!(compvalue<BIGENDIAN(pcompare->compvalue)))
               and_cond=0;
          }
          if (pcompare->comptyp==EQUAL_OR_GREATER){
             if (!(compvalue>=BIGENDIAN(pcompare->compvalue)))
               and_cond=0;
          }
          if (pcompare->comptyp==EQUAL_OR_SMALLER){
             if (!(compvalue<=BIGENDIAN(pcompare->compvalue)))
               and_cond=0;
          }
        }
        // Offsets nachziehen
        and_offset+=SIZE_COMPARE;
        or_offset+=SIZE_COMPARE;
        offset+=SIZE_COMPARE;
      }
      
      // waren alle UND Verknuepfungen WAHR, so ist die ODER Verknuepfung wahr
      if (and_cond){
        or_cond=1;
        //Wurde ein TastenCLICK in eine Verknuepfung einbezogen und ausgewertet, so
        //wird dies gemerkt, und das nachfolgende RELEASE Event unterdrueckt
        if (keyclickdetected)
          if (keyclicknr<MAX_KEYS)
            keys[keyclicknr].clickdetected=1;
      }

      // Offsets nachziehen
      offset+=SIZE_OR_COMPARE;
      or_offset+=SIZE_OR_COMPARE;
    }

    // Es ist ein Relais zu setzen
    if (paction->object.obj_typ==RELAY){
      if (paction->object.obj_nr<MAX_RELAYS)
        relays[paction->object.obj_nr].active=or_cond;
    }
    
    // Es ist ein Zustand zu aendern
    if ((paction->object.obj_typ==STATE) && (or_cond) && (!statemachines[paction->object.obj_nr].has_changed)){
      if (paction->object.obj_nr<MAX_STATEMACHINES){
        statemachines[paction->object.obj_nr].state=BIGENDIAN(paction->val);
        statemachines[paction->object.obj_nr].has_changed=1;
      }
    }

    // Offsets nachziehen
    offset+=SIZE_ACTION;
  }

}

